unit haupt;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls,
  StdCtrls, LCLIntf, LCLType, Menus;

const
  fFenster=clGreen; fFeld=clOlive; fFeldlinien=clBlack;
  fMarkeMarkiert=clYellow; fMarkeNormal=clBlack;
  pOben:TPoint=(X:0;Y:-1); pLinks:TPoint=(X:-1;Y:0);
  pUnten:TPoint=(X:0;Y:1); pRechts:TPoint=(X:1;y:0);
  pRicht:array[0..3]of TPoint=((X:0;Y:-1),(X:-1;Y:0),(X:0;Y:1),(X:1;y:0));
  pNone:TPoint=(X:-1;Y:-1);

type
  TSpielfeld = class;
  TStein = class;
  TSteinfeld = array[0..10,0..11] of TStein;
  { TStein }
  TStein = class
  constructor Create(AOwner:TComponent;a:TPoint);
  private
    Augen:TPoint; // identisch mit Indices in AllSteine; immer X <= Y
    KoordAuslage:TPoint; // immer senkrecht; bleiben konstant während des Spiels;
    KoordSpielfeld:TPoint; // ist pNone, wenn nicht in Spielfeld; ändert sich bei jedem Zug
                           // ist die Spielfeldkoordinate des links/oben liegenden Steinfelds
    WaagrSpielfeld:boolean; // true=waagrecht; false=senkrecht - phys.Lage im Spieleld
    AugenSpielfeld:boolean; // true=AugenX links bzw.oben; false=AugenY links bzw.oben
    KoordLoesung:TPoint;
    WaagrLoesung:boolean;
    AugenLoesung:boolean;
    inAuslage:boolean;
    inLoesung: boolean;
    function inSpielfeld : boolean;
    procedure Zeichne(c:TCanvas;x,y:integer;w,r:boolean);
    procedure EntferneAusSpielfeld;
    procedure SetzInSpielFeld(p:TPoint;w,r:boolean);
    procedure SetzInAuslage;
  public
  end;

  { TMarke }
  TMarke = class(TLabel)
  constructor Create(AOwner:TComponent); override;
  procedure MMouseDown(Sender:TObject;Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
  procedure MMouseMove(Sender:TObject;Shift:TShiftState;X,Y:Integer);
  private
    Pwert:integer;
    Markiert:boolean;
    procedure SetMarkierung(m:boolean);
    procedure SetWert(w:integer);
  public
    property Markierung:boolean Read Markiert Write SetMarkierung;
    property Wert:integer Read Pwert Write SetWert;
  end;

  { TAuswahl }
  TAuswahl = class(TBevel)
  constructor Create(AOwner:TComponent); override;
  procedure AMouseMove(Sender:TObject;Shift:TShiftState;X,Y:Integer);
  procedure AMouseDown(Sender:TObject;Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
  procedure AClick(Sender:TObject);
  procedure ADblClick(Sender:TObject);
  private
    Koord:TPoint;
    Richt:integer; //0=hoch,1=links,2=unten,3=rechts
    function SetKoord(p:TPoint;r:integer;u:boolean):boolean;
  end;

  { TAuslage }
  TAuslage = class
  constructor Create(AOwner:TComponent);
  const
    SteineMassVorgabe:array[1..9]of TPoint = // Breite,Höhe in Steine, nicht Felder
      ((X:2;Y:2),(X:4;Y:2),(X:4;Y:3),(X:6;Y:3),(X:6;Y:4),(X:8;Y:4),(X:8;Y:5),(X:10;Y:5),(X:10;Y:6));
  private
   // Panel:TPanel;
    Steine:array[0..10,0..6]of TStein; // (alle möglichen) Steine in der Auslage
    SteineMass:TRect; // Bereichsgrenzen (inclusive)
  procedure Ini; // bestimme Auslage entsprechend aktuellem MaxPunkte
  public
  end;

  { TSpielfeld }
  TSpielfeld = class
  constructor Create(AOwner:TComponent);
  private
    Panel:TPanel;
    Steine:array[0..10,0..11]of TStein; // (alle möglichen) Steine im Spielfeld
    SteineMass:TRect; // Bereichsgrenzen (inclusive)
    Marken:array[0..10,0..11] of TMarke; // alle Marken
    Besetzt:array[0..10,0..11]of boolean; // false, wenn unbesetzt
    function isBesetzt(p:TPoint):boolean; // zum wahlweisen Aufruf mit Point statt x,y
    procedure Ini; // bestimme Spielfeld zufällig entsprechend aktuellem MaxPunkte
    procedure Markiere(a:TPoint); // markiere alle möglichen Vorkommen mit diesen Augenzahlen
    procedure MarkenUnmarkiert; // alle Markierungen (Felder zeigen mit ...) entfernen
  end;

  { THauptfenster }
  THauptfenster = class(TForm)
    FertigLabel: TLabel;
    AusDateiLabel: TLabel;
    AnzSteineLabel: TLabel;
    AnzAugenLabel: TLabel;
    AnzZwischenLabel: TLabel;
    AnzRueckLabel: TLabel;
    KeineHilfeLabel: TLabel;
    FertigWeiterLabel: TLabel;
    LoesungHinweis: TLabel;
    LoesungUeberschrift: TLabel;
    LoesungWeiter: TLabel;
    FertigPanel: TPanel;
    SpielSpeichernDialog: TSaveDialog;
    SpielLadenDialog: TOpenDialog;
    SpielfeldPanel: TPanel;
    AuslagePanel: TPanel;
    LoesungPanel: TPanel;
    AktivPanel:TPanel; // eines der 3: Auslage, Loesung, Fertig
    HauptMenu: TMainMenu;
    // MenuSpiel
    MenuSpiel: TMenuItem;
    MenuSpielNeu: TMenuItem;
    MenuSpielMit: TMenuItem;
      MenuSpielMit1: TMenuItem;
      MenuSpielMit2: TMenuItem;
      MenuSpielMit3: TMenuItem;
      MenuSpielMit4: TMenuItem;
      MenuSpielMit5: TMenuItem;
      MenuSpielMit6: TMenuItem;
      MenuSpielMit7: TMenuItem;
      MenuSpielMit8: TMenuItem;
      MenuSpielMit9: TMenuItem;
    MenuSpielWieder: TMenuItem;
    MenuSpielTrenn1: TMenuItem;
    MenuSpielLaden: TMenuItem;
    MenuSpielSpeichern: TMenuItem;
    MenuSpielTrenn2: TMenuItem;
    MenuSpielAblegen: TMenuItem;
    MenuSpielHolen: TMenuItem;
    MenuSpielTrenn3: TMenuItem;
    MenuSpielFelder: TMenuItem;
      MenuSpielFeld0: TMenuItem;
      MenuSpielFeld1: TMenuItem;
      MenuSpielFeld2: TMenuItem;
      MenuSpielFeld3: TMenuItem;
      MenuSpielFeld4: TMenuItem;
      MenuSpielFeld5: TMenuItem;
      MenuSpielFeld6: TMenuItem;
      MenuSpielFeld7: TMenuItem;
      MenuSpielFeld8: TMenuItem;
      MenuSpielFeld9: TMenuItem;
    MenuSpielLoesung: TMenuItem;
    MenuSpielTrenn4: TMenuItem;
    MenuSpielEnde: TMenuItem;
    // Menü Hilfe
    MenuHilfe: TMenuItem;
    MenuHilfeRegeln: TMenuItem;
    MenuHilfeBedienung: TMenuItem;
    MenuHilfeHilfestellungen: TMenuItem;
    MenuHilfeStarten: TMenuItem;
    MenuHilfeTrenn1: TMenuItem;
    MenuHilfeSpiel: TMenuItem;
    MenuHilfeTrenn2: TMenuItem;
    MenuHilfeInformation: TMenuItem;
    procedure FormActivate(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormClick(Sender: TObject);
    procedure AuslagePanelMouseDown(Sender:TObject;Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
    procedure AuslagePanelPaint(Sender: TObject);
    procedure MenuHilfeBedienungClick(Sender: TObject);
    procedure MenuHilfeHilfestellungenClick(Sender: TObject);
    procedure MenuHilfeInformationClick(Sender: TObject);
    procedure MenuHilfeRegelnClick(Sender: TObject);
    procedure MenuHilfeSpielClick(Sender: TObject);
    procedure MenuHilfeStartenClick(Sender: TObject);
    procedure SpielfeldPanelMouseDown(Sender:TObject;Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
    procedure SpielfeldPanelMouseMove(Sender:TObject;Shift:TShiftState;X,Y:Integer);
    procedure SpielfeldPanelPaint(Sender: TObject); // Linien ins SpielfeldPanel zeichnen
    procedure SpielfeldPanelResize(Sender: TObject);
    procedure LoesungPanelPaint(Sender: TObject);
    procedure MenuSpielNeuClick(Sender: TObject);
    procedure MenuSpielMitClick(Sender: TObject); // siehe auch MenuSpielMitItemClick
    procedure MenuSpielWiederClick(Sender: TObject);
    procedure MenuSpielLadenClick(Sender: TObject);
    procedure MenuSpielSpeichernClick(Sender: TObject);
    procedure MenuSpielAblegenClick(Sender: TObject);
    procedure MenuSpielHolenClick(Sender: TObject);
    procedure MenuSpielFelderClick(Sender: TObject); // siehe auch MenuSpielFelderItemClick
    procedure MenuSpielLoesungClick(Sender: TObject);
    procedure MenuSpielEndeClick(Sender: TObject);
  private
    procedure MaxPunkteNeu(mp:integer);
    procedure Ini(mp:integer;res:boolean=false);
    procedure PanelShow(pan:TPanel);
    procedure MenuSpielMitItemClick(Sender: TObject); // Click für die einzelnen "Spiel mit"-Werte
    procedure MenuSpielFelderItemClick(Sender: TObject); // Click für die einzelnen "Felder zeigen"-Werte
    procedure Blockieren(b:boolean);
    procedure SpielSteineZeichnen(pan:TPanel;feld:TSteinfeld);
    function SpielToString(f:boolean):string;
    function TesteSpielstring(z:string;f:boolean):integer;
    procedure StringToSpiel(z:String;f:boolean);
    procedure SpielFertig;
  end;

var
  Hauptfenster: THauptfenster;
  Auswahl:TAuswahl;
  Auslage:TAuslage;
  Spielfeld:TSpielfeld;
  MaxPunkte:integer;
  SteinBreite:integer;
  AllSteine:array[0..9,0..9] of TStein; // alle Steine im Spiel
  AnzAugen,AnzSteine,AnzRueck,AnzZwischen:integer; {Hilfen benutzt}
  AusDatei:boolean;
  ArbeitsVerzeichnis:string;
  Zwischenstand:string;
  Block:boolean;
  Ende:boolean;

implementation
uses hilfe, Dos, Process;

{$R *.lfm}

{ THauptfenster }

procedure THauptfenster.FormCreate(Sender: TObject);
var a1,a2:integer; s:TStein; qvz,cr,os:string; f:file; i:integer;
    bsp:array[1..9]of string=('Eins','Zwei','Drei','Vier','Fuenf','Sechs','Sieben','Acht','Neun');
begin
  Ende:=false;
  qvz:=ExtractFileDir(ParamStr(0)); // Quellverzeichnis von Installation
  os:=''; // Linux oder Windows, wenn nicht portable Version
  if DirectoryExists(qvz+'/Beispiele') then // installierte Version
  begin
    if GetEnvironmentVariable('HOME')<>'' then // Linux
      begin os:='Linux'; ArbeitsVerzeichnis:=GetEnvironmentVariable('HOME')+'/.domino'; end
    else if GetEnvironmentVariable('APPDATA')<>'' then // Windows
      begin os:='Windows'; ArbeitsVerzeichnis:=GetEnvironmentVariable('APPDATA')+'\Domino'; end;
    if os<>'' then
      if Not DirectoryExists(ArbeitsVerzeichnis) then // Arbeitsverzeichnis erzeugen
        begin
          if CreateDir(ArbeitsVerzeichnis) then // Beispieldateien ins Arbeitsverzeichnis kopieren
          begin
            for i:=1 to 9 do
            begin
              CopyFile(qvz+'/Beispiele/'+bsp[i]+'.dom',Arbeitsverzeichnis+'/'+bsp[i]+'.dom',true);
              if os='Linux' then
                RunCommand('/bin/bash',['-c','chmod a-w '+Arbeitsverzeichnis+'/'+bsp[i]+'.dom'],cr)
              else
                begin
                  AssignFile(f,Arbeitsverzeichnis+'/'+bsp[i]+'.dom');
                  SetFAttr(f,ReadOnly);
                end;
            end;
          end else os:='';
        end;
  end;
  if os='' then ArbeitsVerzeichnis:=ExtractFileDir(ParamStr(0));
  SpielLadenDialog.InitialDir:=ArbeitsVerzeichnis;
  SpielSpeichernDialog.InitialDir:=ArbeitsVerzeichnis;
  for a1:=0 to 9 do for a2:=0 to 9 do
  begin // Steine erzeugen
    if a1<=a2 then s:=TStein.Create(Self,Point(a1,a2)) else s:=AllSteine[a2,a1];
    AllSteine[a1,a2]:=s;
  end;
  Randomize;
  Auslage:=TAuslage.Create(Self);
  Spielfeld:=TSpielfeld.Create(Self);
  Ini(6); // Start immer mit 6-er-Domino
end;

procedure THauptfenster.FormClose(Sender:TObject;var CloseAction:TCloseAction);
begin Block:=true; Ende:=true; end;

procedure THauptfenster.FormActivate(Sender: TObject);
begin AktivPanel.Repaint; end; // Sonst erscheint anfangs nicht die Auslage

procedure THauptfenster.FormResize(Sender: TObject);
begin Ini(0,true); end;

procedure THauptfenster.Ini(mp:integer;res:boolean=false);
  // mp=MaxPunkte, 0 wenn nicht ändern; neu=Spielfeld neu belegen; res=aus Resize (Kein Loesung-Hide)
var fbreit,fhoch,abreit,xrechts,yoben,abstand:integer;
begin
  if mp>0 then MaxPunkteNeu(mp);
  SteinBreite:=ClientWidth div (Spielfeld.SteineMass.Width+Auslage.SteineMass.Width+10);
    // AuslagePanel ist immer >= SpielfeldPanel
    // Feldabstände mindestens SteinBreite; jetzt erweitern auf gleichmäßig
  fbreit:=(Spielfeld.SteineMass.Width+2)*SteinBreite;
  fhoch:=(Spielfeld.SteineMass.Height+2)*Steinbreite;
  abreit:=(Auslage.SteineMass.Width+2)*SteinBreite;
  abstand:=(ClientWidth-fbreit-abreit) div 3; // Abstand links, rechts, Mitte
  xrechts:=fbreit+2*abstand;
  yoben:=(ClientHeight-fhoch) div 2;
  SpielfeldPanel.SetBounds(abstand,yoben,fbreit,fhoch); // SpielfeldPanel ist links
  AuslagePanel.SetBounds(xrechts,round((ClientHeight-(2*Auslage.SteineMass.Height+2)*SteinBreite)/2),
    abreit,(2*Auslage.SteineMass.Height+2)*SteinBreite); // AuslagePanel ist rechts, alternativ Lösung
  LoesungPanel.SetBounds(xrechts,yoben,fbreit,round(fhoch*1.2)); // Lösung ist recht, alternativ AuslagePanel
  LoesungUeberschrift.Font.Size:=fhoch div 30;
  LoesungHinweis.Font.Size:=fhoch div 36;
  LoesungWeiter.Font.Size:=fhoch div 36;
  FertigPanel.SetBounds(xrechts,round((ClientHeight-FertigPanel.Height)/2),FertigPanel.Width,FertigPanel.Height);
  if not res then
  begin
    Auslage.Ini; // AuslagePanel entsprechend Punktzahl
    Spielfeld.Ini; // Spielfeld random belegen
    Blockieren(false);
    PanelShow(AuslagePanel);
  end;
end;

procedure THauptfenster.PanelShow(pan:TPanel);
var p:TPanel; pa:array[0..2] of TPanel;
begin
  if Ende then exit;
  pa[0]:=AuslagePanel; pa[1]:=LoesungPanel; pa[2]:=FertigPanel;
  for p in pa do
    if p=pan then p.Show else p.Hide;
  AktivPanel:=pan;
  Spielfeld.MarkenUnmarkiert;
  Activate;
end;

procedure THauptfenster.FormClick(Sender: TObject);
begin Spielfeld.MarkenUnmarkiert; end; // bei Klick aufs Fenster Markierungen entfernen

procedure THauptfenster.SpielfeldPanelPaint(Sender: TObject);
var i:integer;
begin // Linien ins SpielfeldPanel zeichnen
  if Ende then exit;
  SpielfeldPanel.Canvas.Brush.Color:=fFeld; SpielfeldPanel.Canvas.Pen.Color:=fFeldlinien;
  SpielfeldPanel.Canvas.FillRect(Rect(1,1,SpielfeldPanel.Width,SpielfeldPanel.Height));
  for i:=0 to Spielfeld.SteineMass.Height do
  begin
     SpielfeldPanel.Canvas.MoveTo(SteinBreite,(i+1)*SteinBreite);
     SpielfeldPanel.Canvas.LineTo((Spielfeld.SteineMass.Width+1)*SteinBreite,(i+1)*SteinBreite);
  end;
  for i:=0 to Spielfeld.SteineMass.Width do
  begin
     SpielfeldPanel.Canvas.MoveTo((i+1)*SteinBreite,SteinBreite);
     SpielfeldPanel.Canvas.LineTo((i+1)*SteinBreite,(Spielfeld.SteineMass.Height+1)*SteinBreite);
  end;
  SpielSteineZeichnen(SpielfeldPanel,Spielfeld.Steine);
end;

procedure THauptfenster.SpielfeldPanelResize(Sender: TObject);
var d,x,y:integer;
begin
  d:=SteinBreite div 2 + 1; // Marken etwa 1/2 so gross wie Steine
  for x:=0 to Spielfeld.SteineMass.Width-1 do for y:=0 to Spielfeld.SteineMass.Height-1 do // Marken platzieren
  begin
    Spielfeld.Marken[x,y].SetBounds((x+1)*SteinBreite+(SteinBreite-d)div 2+1,
                                    (y+1)*SteinBreite+(SteinBreite-d)div 2,d,d);
    Spielfeld.Marken[x,y].Font.Height:=d-1;
  end;
end;

procedure THauptfenster.LoesungPanelPaint(Sender: TObject);
var a1,a2:integer; LoesungSteine:TSteinfeld; s:TStein;
begin
  if Ende then exit;
  for a1:=0 to MaxPunkte do for a2:=a1 to MaxPunkte do
  begin // alle Steine ins Lösungsfeld
    s:=AllSteine[a1,a2];
    LoesungSteine[s.KoordLoesung.X,s.KoordLoesung.Y]:=s;
    if s.WaagrLoesung then LoesungSteine[s.KoordLoesung.X+1,s.KoordLoesung.Y]:=s
      else LoesungSteine[s.KoordLoesung.X,s.KoordLoesung.Y+1]:=s;
  end;
  SpielSteineZeichnen(LoesungPanel,LoesungSteine);
end;

procedure THauptfenster.SpielSteineZeichnen(pan:TPanel;feld:TSteinfeld);
var x,y,n:integer; w,a,loe,fertig,ok:boolean;  p:TPoint; s:TStein;
    gezeichnet:array[0..10,0..11] of boolean;
    nbw,nbs,nb:array[0..2]of TPoint; tb:TPoint;
begin
  // Lösungsfeld hat dieselben Maße wie Spielfeld
  for x:=0 to Spielfeld.SteineMass.Width-1 do
    for y:=0 to Spielfeld.SteineMass.Height-1 do gezeichnet[x,y]:=false;
  nbw[0]:=Point(0,-1); nbw[1]:=Point(1,-1); nbw[2]:=Point(2,0);
  nbs[0]:=Point(1,0); nbs[1]:=Point(1,1); nbs[2]:=Point(0,-1);
  fertig:=false;
  while (not fertig) do
  begin
  fertig:=true;
  for x:=Spielfeld.SteineMass.Width-1 downto 0 do
    for y:=0 to Spielfeld.SteineMass.Height-1 do
      if not gezeichnet[x,y] then
      begin  // Spielsteine platzieren, von rechts nach links, von oben nach unten
        ok:=true;
        s:=feld[x,y];
        if s<>nil then // vorhandener Stein
        begin
          if pan=SpielfeldPanel then
          begin
            loe:=false;
            p:=s.KoordSpielfeld; // dadurch evtl. korrigiert
            w:=s.WaagrSpielfeld;
            a:=s.AugenSpielfeld;
          end else
          begin
            loe:=true;
            p:=s.KoordLoesung;
            w:=s.WaagrLoesung;
            a:=s.AugenLoesung;
          end;
          if w then nb:=nbw else nb:=nbs;
          for n:=0 to 2 do // teste Nachbarumgebung
          begin
            tb:=p+nb[n]; // teste Nachbarpunkt (für Loesung ist Besetzt... irrelevant)
            if PtInRect(Spielfeld.SteineMass,tb) and
              (loe or Spielfeld.Besetzt[tb.X,tb.Y]) and not gezeichnet[tb.X,tb.Y] then
              begin ok:=false; break; end;
          end;
          if ok then
          begin
            s.Zeichne(pan.Canvas,(p.X+1)*SteinBreite,(p.Y+1)*SteinBreite,w,a);
            gezeichnet[p.X,p.Y]:=true; if w then gezeichnet[p.X+1,p.Y]:=true else gezeichnet[p.X,p.Y+1]:=true;
            fertig:=false; // solange noch Steine gezeichnet wurden
          end;
        end;
      end;
  end;
  end;

procedure THauptfenster.SpielfeldPanelMouseDown(Sender:TObject;Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
var maus:TPoint; s:TStein;
begin
  if Block then exit;
  Spielfeld.MarkenUnmarkiert;
  maus:=Point(X div SteinBreite -1,Y div SteinBreite -1); // akt MausKoord. in Steineinheiten
  if not PtInRect(Spielfeld.SteineMass,maus) then exit; // außerhalb
  // innerhalb ist nur auf Stein (sonst Auslage: eigene Mausereignisse)
  s:=Spielfeld.Steine[maus.X,maus.Y]; if s=nil then exit;
  if Button=mbLeft then
  begin // zeige weitere Möglichkeiten für diesen Stein
    Spielfeld.Markiere(s.Augen);
  end;
  if Button=mbRight then
  begin // Stein in Vorrat zurücklegen
    s.EntferneAusSpielfeld;
    s.SetzInAuslage;
    AnzRueck:=AnzRueck+1;
    SpielfeldPanel.Repaint; AuslagePanel.Repaint;
  end;
end;

procedure THauptfenster.SpielfeldPanelMouseMove(Sender:TObject;Shift:TShiftState;X,Y:Integer);
var maus:TPoint; b:boolean;
begin
  if Block then exit;
  maus:=Point(X div SteinBreite -1,Y div SteinBreite -1); // akt MausKoord. in Steineinheiten
  if not PtInRect(Spielfeld.SteineMass,maus) then
    begin Auswahl.Visible:=false; exit; end; // auf Rand: Auswahl verbergen
  if Spielfeld.Besetzt[maus.X,maus.Y] then begin Auswahl.Visible:=false; exit; end; // auf Stein
  if(maus=Auswahl.Koord)and(Auswahl.Visible) then exit; // keine Änderung
  b:=Auswahl.SetKoord(maus,Auswahl.Richt,true);
  if b then Auswahl.Visible:=true;
  //Application.ProcessMessages;
end;

procedure THauptfenster.AuslagePanelPaint(Sender: TObject);
var a1,a2:integer;
procedure gzeichne(a1,a2:integer);
var s:TStein;
begin
  if Ende then exit;
  s:=AllSteine[a1,a2];
  if s.inAuslage then
    s.Zeichne(AuslagePanel.Canvas,
      (s.KoordAuslage.X+1)*SteinBreite,(2*s.KoordAuslage.Y+1)*SteinBreite,false,true);
end;
begin
  for a1:=0 to Auslage.SteineMass.Height-1 do for a2:=MaxPunkte downto a1 do gzeichne(a1,a2);
  for a1:=MaxPunkte downto Auslage.SteineMass.Height do for a2:=MaxPunkte downto a1 do gzeichne(a1,a2);
end;

procedure THauptfenster.AuslagePanelMouseDown(Sender:TObject;Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
var ix,iy:integer; s:TStein;
begin
  if Block then exit;
  if Button<>mbLeft then exit;
  Spielfeld.MarkenUnmarkiert;
  ix:=X div Steinbreite -1; iy:=Y div SteinBreite-1; // Steinkoordinaten (iy wegen Abfrage noch *2)
  if iy>=0 then iy:=iy div 2;
  if not PtInRect(Auslage.SteineMass,Point(ix,iy)) then exit;  // Klick nicht im Feld
  s:=Auslage.Steine[ix,iy]; if s=nil then exit;
  if not s.inAuslage then exit;
  Spielfeld.Markiere(s.Augen);
end;

procedure THauptfenster.MaxPunkteNeu(mp:integer);
var i:integer; v:TPoint;
begin
  if Ende then exit;
  MaxPunkte:=mp;
  Spielfeld.SteineMass:=Rect(0,0,MaxPunkte+1,MaxPunkte+2);
  v:=Auslage.SteineMassVorgabe[MaxPunkte]; // Vorgabe für Auslage-Masse
  Auslage.SteineMass:=Rect(0,0,v.X,v.Y); // AuslagePanel anpassen
  for i:=0 to 9 do // nur "Felder zeigen" mit gültigen Werten
  begin
    MenuSpielFelder.Items[i].Visible:=(i<=MaxPunkte);
    MenuSpielFelder.Items[i].Checked:=false;
  end;
end;

procedure THauptfenster.MenuSpielNeuClick(Sender: TObject);
begin Ini(0); end;

procedure THauptfenster.MenuSpielMitClick(Sender: TObject);
var i:integer;
begin // an Untermenüpunkt weiterreichen
  for i:=0 to 8 do MenuSpielMit.Items[i].OnClick:=@MenuSpielMitItemClick;
end;

procedure THauptfenster.MenuSpielMitItemClick(Sender: TObject);
var i,mp:integer;
begin
     if TryStrToInt(TMenuItem(Sender).Caption,mp) then
     begin
        for i:=0 to 8 do MenuSpielMit.Items[i].Checked:=false;
        TMenuItem(Sender).Checked:=true;
        Ini(mp);
     end;
end;

procedure THauptfenster.MenuSpielWiederClick(Sender: TObject);
var x,y:integer;
begin
  for x:=0 to Spielfeld.SteineMass.Width-1 do for y:=0 to Spielfeld.SteineMass.Height-1 do
  begin
    Spielfeld.Besetzt[x,y]:=false;
    if Spielfeld.Steine[x,y]<>nil then
      begin Spielfeld.Steine[x,y].KoordSpielfeld:=pNone; Spielfeld.Steine[x,y]:=nil; end;
    Spielfeld.Marken[x,y].Visible:=true;
  end;
  SpielfeldPanel.Repaint;
  Auslage.Ini;
  PanelShow(AuslagePanel);
  Blockieren(false);
end;

procedure THauptfenster.MenuSpielLadenClick(Sender: TObject);
var s:TStringList; i,mp:integer; fn,fm:string;
begin
  SpielLadenDialog.FileName:=''; fm:=''; mp:=MaxPunkte;
  if SpielLadenDialog.Execute then
  begin
    s:=TStringList.Create;
    fn:=SpielLadenDialog.FileName;
    try
      s.LoadFromFile(fn);
    except
      on E:exception do fm:='Datei '+fn+lineending+'ist nicht vorhanden oder fehlerhaft';
    end;
    if fm='' then
    begin
      if(s.Count=0)or(s.Count>2) then mp:=-1 else mp:=TesteSpielstring(s[0],false);
      if (mp>=0)and(s.Count=2) then
        begin if length(s[1])=length(s[0]) then mp:=TesteSpielstring(s[1],true) else mp:=-1; end;
      if mp<0 then fm:=fn+lineending+'ist keine fehlerfreie Domino-Datei';
    end;
    if fm='' then
    begin
      Ini(mp);
      StringToSpiel(s[0],false); // erst die Lösung (wird für Spielfeld-Marken gebraucht)
      if s.Count=2 then
        begin StringToSpiel(s[1],true); AusDatei:=true end; // Spielstand, wenn vorhanden
      for i:=0 to 8 do MenuSpielMit.Items[i].Checked:=(i=MaxPunkte-1);
    end;
    if fm<>'' then showmessage(fm);
    s.Free;
    SpielfeldPanel.Repaint;
  end;
end;

procedure THauptfenster.MenuSpielSpeichernClick(Sender: TObject);
var s:TStringList; sz,fn:string; ok:boolean;
begin
  SpielSpeichernDialog.FileName:='';
  ok:=SpielSpeichernDialog.Execute; // hier bereits System-Fehlermeldung bei Schreibschutz
  if ok then
  begin
    s:=TStringList.Create;
    s.Add(SpielToString(false)); // Lösung
    sz:=SpielToString(true);
    if sz<>'' then s.Add(sz); // gelegte Steine, wenn vorhanden
    ok:=false;
    fn:=SpielSpeichernDialog.FileName;
    if fn<>'' then
    try
      s.SaveToFile(fn);
    except
      on E:Exception do
        showmessage('Speichern der Datei geht nicht:'+lineending+SpielSpeichernDialog.FileName);
    end;
    s.Free;
   end;
end;

procedure THauptfenster.MenuSpielAblegenClick(Sender: TObject);
begin Zwischenstand:=SpielToString(true); end;

procedure THauptfenster.MenuSpielHolenClick(Sender: TObject);
var z:string;
begin
   if Zwischenstand<>'' then
   begin
     z:=Zwischenstand; Ini(0); Zwischenstand:=z; // wird beim Ini '' gesetzt
     StringToSpiel(Zwischenstand,true);
     AnzZwischen:=AnzZwischen+1;
   end;
end;

function THauptfenster.SpielToString(f:boolean):string;
  {f=true:akt.Stellung; f=false:Lösung
   Koordinaten waag,senk: A,a=0;  B,b=1; ...
   1.Buchstabe (waag): groß(A...):senkrecht, klein(a...):waagrecht;
   2.Buchstabe (senk): groß(A...):a1>a2; klein(a...):a1<=a2
   Reihenfolge 0-0,0-1,0-2,...,1-1,1-2,...}
var a1,a2,a,b:integer; z:string; s:TStein; k:TPoint; w:boolean;
begin
  z:='';
  for a1:=0 to MaxPunkte do for a2:=a1 to MaxPunkte do
  begin
    s:=AllSteine[a1,a2]; k:=pNone;
    if not f then // Lösung
      begin k:=s.KoordLoesung; w:=s.WaagrLoesung end
    else // Spielfeld
      begin if s.inSpielfeld then begin k:=s.KoordSpielfeld; w:=s.WaagrSpielfeld; end end;
    if k.X>=0 then
    begin
      if w then a:=ord('a') else a:=ord('A');
      if a1=Spielfeld.Marken[k.X,k.Y].Pwert then b:=ord('a') else b:=ord('A');
      z:=z+chr(k.x+a)+chr(k.y+b);
    end else z:=z+'00';
  end;
  Result:=z;
end;

function THauptfenster.TesteSpielstring(z:string;f:boolean):integer;
// Ergebnis: Wert für MaxPunkte; -1 wenn ungültig
var i,j,mp,x,y:integer; mx:real; p1,p2:TPoint; w:boolean;
    sb:array[0..10,0..11]of boolean;
begin
  z:=trim(z); i:=1; result:=-1;
  for i:=0 to 10 do for j:=0 to 11 do sb[i,j]:=false; // Belegt-Kontrolle
  mx:=(sqrt(4*length(z)+1)-3)/2; // MaxPunkte aus Ablage-Länge berechnen
  if(frac(mx)<>0)or(mx<1) then exit; // ungültig
  mp:=round(mx);
  for i:=1 to (length(z)div 2) do
  begin
    if z[2*i-1]='0' then
    begin
      if (z[2*i]<>'0')or(not f) then exit; // unbelegt nur bei Spielfeld
    end else
    begin
      x:=ord(z[2*i-1])-ord('a');  w:=x>=0;
      if x<0 then x:=ord(z[2*i-1])-ord('A');
      y:=ord(z[2*i])-ord('a'); if y<0 then y:=ord(z[2*i])-ord('A');
      p1:=Point(x,y);
      if w then p2:=p1+pRechts else p2:=p1+pUnten;
      if not PtInRect(Rect(0,0,mp+1,mp+2),p1) or not PtInRect(Rect(0,0,mp+1,mp+2),p2)
         then exit; // Koordinate außerhalb Spielfeld
      if sb[x,y] then exit; // Feld doppelt besetzt
    end;
  end;
  result:=mp;
end;

procedure THauptfenster.StringToSpiel(z:string;f:boolean);
var i,a1,a2,b1,b2,x,y:integer; p1,p2:TPoint; s:TStein; w,v:boolean;
begin
  z:=trim(z); i:=1;
  for a1:=0 to MaxPunkte do for a2:=a1 to MaxPunkte do
  begin
    w:=(z[i]>='a'); if w then x:=ord('a') else x:=ord('A');
    v:=(z[i+1]>='a'); if v then y:=ord('a') else y:=ord('A');
    if z[i]='0' then begin x:=-1; y:=-1 end
                else begin x:=ord(z[i])-x; y:=ord(z[i+1])-y end;
    s:=AllSteine[a1,a2];
    if v then begin b1:=a1; b2:=a2; end else begin b1:=a2; b2:=a1; end;
    p1:=Point(x,y);
    if w then p2:=p1+pRechts else p2:=p1+pUnten;
    if not f then // Lösung
    begin
      s.KoordLoesung:=p1;
      s.WaagrLoesung:=w;
      s.AugenLoesung:=v;
      Spielfeld.Marken[p1.X,p1.Y].SetWert(b1);
      Spielfeld.Marken[p2.X,p2.Y].SetWert(b2);
    end else // Spielfeld
    begin
      if x>=0 then
      begin
        s.SetzInSpielfeld(Point(x,y),w,true);
      end;
    end;
    i:=i+2;
  end;
end;

procedure THauptfenster.MenuSpielFelderClick(Sender: TObject);
var i:integer;
begin // an Untermenüpunkt weiterreichen
   for i:=0 to 9 do MenuSpielFelder.Items[i].OnClick:=@MenuSpielFelderItemClick;
end;

procedure THauptfenster.MenuSpielFelderItemClick(Sender: TObject);
var i,f,x,y:integer;
begin
  if TryStrToInt(TMenuItem(Sender).Caption,f) then
  begin // Felder mit gewünschter Augenzahl markieren
    for i:=0 to 9 do MenuSpielFelder.Items[i].Checked:=(i=f);
    for x:=0 to 10 do for y:=0 to 11 do
      Spielfeld.Marken[x,y].SetMarkierung(Spielfeld.Marken[x,y].Pwert=f);
    AnzAugen:=AnzAugen+1;
  end;
end;

procedure THauptfenster.MenuSpielLoesungClick(Sender: TObject);
begin
  Blockieren(true);
  MenuSpielLoesung.Enabled:=false;
  PanelShow(LoesungPanel);
end;

procedure THauptfenster.SpielFertig;
begin
  Blockieren(true);
  AusDateiLabel.Visible:=AusDatei;
  AnzSteineLabel.Caption:=inttostr(AnzSteine)+' mal Steine angezeigt';
  AnzSteineLabel.Visible:=AnzSteine>0;
  AnzAugenLabel.Caption:=inttostr(AnzAugen)+' mal Augen angezeigt';
  AnzAugenLabel.Visible:=AnzAugen>0;
  AnzZwischenLabel.Caption:=inttostr(AnzZwischen)+' mal Zwischenstand geholt';
  AnzZwischenLabel.Visible:=AnzZwischen>0;
  AnzRueckLabel.Caption:=inttostr(AnzRueck)+' mal Steine korrigiert';
  AnzRueckLabel.Visible:=AnzRueck>0;
  KeineHilfeLabel.Visible:=(AnzSteine+AnzAugen+AnzZwischen+AnzRueck=0)and not AusDatei;
  PanelShow(FertigPanel);
end;

procedure THauptfenster.MenuSpielEndeClick(Sender: TObject);
begin  Block:=true; Ende:=true; close; end;

procedure THauptfenster.MenuHilfeRegelnClick(Sender: TObject);
begin Hilfefenster.Zeige(0); end;

procedure THauptfenster.MenuHilfeBedienungClick(Sender: TObject);
begin Hilfefenster.Zeige(1); end;

procedure THauptfenster.MenuHilfeHilfestellungenClick(Sender: TObject);
begin Hilfefenster.Zeige(2); end;

procedure THauptfenster.MenuHilfeStartenClick(Sender: TObject);
begin   Hilfefenster.Zeige(3); end;

procedure THauptfenster.MenuHilfeSpielClick(Sender: TObject);
begin Hilfefenster.Zeige(4); end;

procedure THauptfenster.MenuHilfeInformationClick(Sender: TObject);
begin Hilfefenster.Zeige(5); end;

procedure THauptfenster.Blockieren(b:boolean);
begin
  Block:=b;
  MenuSpielAblegen.Enabled:=not b;
  MenuSpielHolen.Enabled:=not b;
  if not b then MenuSpielLoesung.Enabled:=true; // nur entblockieren
  MenuSpielFelder.Enabled:=not b;
end;

{ TAuslage }
constructor TAuslage.Create(AOwner:TComponent);
begin
  //Panel:=Hauptfenster.AuslagePanel;
end;

procedure TAuslage.Ini; // bestimme AuslagePanel entsprechend MaxPunkte
var a1,a2,x,y:integer; s:TStein;
begin // Richtung ist immer senkrecht; nur z.B. 0-1, nicht 1-0
  if Ende then exit;
  for x:=0 to 10 do for y:=0 to 6 do Steine[x,y]:=nil; // erst alle unbesetzt
  for a1:=0 to 9 do for a2:=a1 to 9 do
  begin
    s:=AllSteine[a1,a2]; // Stein aus dem Spielvorrat
    s.inAuslage:=false; // Stein liegt (noch) nicht auf der Auslage
    if (a1<=MaxPunkte)and(a2<=MaxPunkte) then // nur gültige Steine
    begin
      if a1<SteineMass.Height then
        begin x:=a2-a1; y:=a1 end
      else // 2.Gruppe daneben, um Platz zu sparen                               
        begin
          if MaxPunkte+1<SteineMass.Width then x:=a2+1 else x:=a2;
          y:=2*SteineMass.Height-a1-1;
        end;
      Steine[x,y]:=s; // (Auslage.Steine) der Stein im Vorrat
      s.KoordAuslage:=Point(x,y); // Auslagekoordinaten in Steineigenschaften
      s.inAuslage:=true;
    end;
    end;
    Hauptfenster.AuslagePanel.Repaint;
end;

{ TSpielfeld }
constructor TSpielfeld.Create(AOwner:TComponent);
var x,y:integer;
begin
  Panel:=Hauptfenster.SpielfeldPanel;
  for x:=0 to 10 do for y:=0 to 11 do
    Marken[x,y]:=TMarke.Create(AOwner); // Marken erzeugen
  Hauptfenster.SpielfeldPanel.BringToFront;
  Auswahl:=TAuswahl.Create(Hauptfenster);
end;

procedure TSpielfeld.Ini; // Spielfeld zufällig mit neuer Stellung belegen
var a,x,y,xm,ym,r,s,f,a1,a2,i:integer; l:boolean; v:array[1..55]of TPoint;
begin
  if Ende then exit;
  for x:=0 to 10 do for y:=0 to 11 do Marken[x,y].SetWert(-1);
    // alle Marken erst einmal ungültig; insbesondere die außerhalb des Spielfelds
  for x:=0 to SteineMass.Width-1 do for y:=0 to SteineMass.Height-1 do
    begin Besetzt[x,y]:=false; Steine[x,y]:=nil; end; // zunächst unbelegt
  a:=0; y:=0;
  for a1:=0 to MaxPunkte do for a2:=a1 to MaxPunkte do
    begin
      AllSteine[a1,a2].KoordSpielfeld:=pNone; // kein Stein im Spielfeld
      a:=a+1; v[a]:=Point(a1,a2); // Steine im Spiel (Augen, nicht grafisch)
    end;
  while a>0 do
  begin
    r:=-1; s:=2*MaxPunkte+4; f:=0; xm:=-1; ym:=-1;
    for x:=0 to SteineMass.Width-1 do
    begin
      for y:=0 to SteineMass.Height-1 do
      begin
        f:=0; // Anzahl freie Möglichkeiten um den Punkt x,y
        if Marken[x,y].Pwert<0 then // Feldpunkt noch frei
        begin
          if x+y<s then begin s:=x+y; xm:=x; ym:=y end;
          if y+1<SteineMass.Height then
            if Marken[x,y+1].Pwert<0 then // nächsthöherer Punkt frei
              begin f:=f+1; r:=1 end; // Richtung senkrecht
            if x+1<SteineMass.Width then
              if Marken[x+1,y].Pwert<0 then // nächstrechter Punkt frei
                 begin f:=f+1; r:=0 end; // Richtung waagrecht
              if y>0 then if Marken[x,y-1].Pwert<0 then f:=f+1; // auch nächstniedrigerer
              if x>0 then if Marken[x-1,y].Pwert<0 then f:=f+1; // auch linker
        end;
        if f=1 then break; // nur eine Möglichkeit
      end;
      if f=1 then break; // nur eine Möglichkeit
    end;
    if f>1 then begin r:=Random(2); x:=xm; y:=ym end; // mehrere: zufällige Richtung
    i:=Random(a)+1; // Stein aus Spiel zufällig auswählen
    if Random(2)>0 then // Lage zufällig auswählen (z.B. 0-1 oder 1-0)
    begin // Augenzahlen des gewählten Steins in FeldMarken übertragen; Lage z.B. 0-1
        Marken[x,y].SetWert(v[i].X); Marken[x+1-r,y+r].SetWert(v[i].Y); l:=true;
    end else
    begin // Lage z.B. 1-0
        Marken[x,y].SetWert(v[i].Y); Marken[x+1-r,y+r].SetWert(v[i].X); l:=false
    end;
      // für Stein im Vorrat Koordinaten auf Lösungsfeld und Richtung eingeben:
      AllSteine[v[i].X,v[i].Y].KoordLoesung:=Point(x,y);
      AllSteine[v[i].X,v[i].Y].WaagrLoesung:=(r=0);
      AllSteine[v[i].X,v[i].Y].AugenLoesung:=l;
      v[i]:=v[a]; a:=a-1; // verwendeter Stein nicht mehr im Spiel
  end;
  MarkenUnmarkiert;
  Auswahl.Richt:=1; // waagrecht als Voreinstellung
  Zwischenstand:='';
  AnzAugen:=0; AnzSteine:=0; AnzRueck:=0; AnzZwischen:=0; AusDatei:=false;
  Panel.Repaint;
end;

function TSpielfeld.isBesetzt(p:TPoint):boolean;
begin result:=Besetzt[p.X,p.Y]; end;

procedure TSpielfeld.Markiere(a:TPoint);
var erfolg:boolean; x,y:integer;
begin
  MarkenUnmarkiert; erfolg:=false;
  for x:=0 to SteineMass.Width-1 do for y:=0 to SteineMass.Height-1 do
  begin
    if not Besetzt[x,y] then
    begin
      if x<SteineMass.Width-1 then if not Besetzt[x+1,y] then
        if ((Marken[x,y].Pwert=a.X)and(Marken[x+1,y].Pwert=a.Y))
           or((Marken[x,y].Pwert=a.Y)and(Marken[x+1,y].Pwert=a.X)) then
          begin // waagrechte Möglichkeit
            Marken[x,y].SetMarkierung(true);
            Marken[x+1,y].SetMarkierung(true);
            erfolg:=true;
          end;
        if y<SteineMass.Height-1 then if not Besetzt[x,y+1] then
          if ((Marken[x,y].Pwert=a.X)and(Marken[x,y+1].Pwert=a.Y))
             or((Marken[x,y].Pwert=a.Y)and(Marken[x,y+1].Pwert=a.X)) then
            begin // senkrechte Möglichkeit
               Marken[x,y].SetMarkierung(true);
               Marken[x,y+1].SetMarkierung(true);
               erfolg:=true;
            end;
    end;
 end;
 if erfolg then AnzSteine:=AnzSteine+1;
end;

procedure TSpielfeld.MarkenUnmarkiert;
var x,y:integer;
begin // alle Markierungen entfernen
  if Ende then exit;
  for x:=0 to 10 do for y:=0 to 11 do Marken[x,y].SetMarkierung(false);
end;

{ TMarke }
constructor TMarke.Create(AOwner:TComponent);
begin
   inherited Create(AOwner);
   Parent:=Hauptfenster.SpielfeldPanel;
   OnMouseDown:=@MMouseDown; OnMouseMove:=@MMouseMove;
   Transparent:=true; Visible:=false;
   Alignment:=taCenter; Font.Style:=[fsBold];
end;

procedure TMarke.MMouseDown(Sender:TObject;Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
begin
  Hauptfenster.SpielfeldPanelMouseDown(Sender,Button,Shift,Left+X,Top+Y);
end;

procedure TMarke.MMouseMove(Sender:TObject;Shift:TShiftState;X,Y:Integer);
begin
  Hauptfenster.SpielfeldPanelMouseMove(Sender,Shift,Left+X,Top+Y);
end;

procedure TMarke.SetMarkierung(m:boolean);
begin // Markierung durch Schriftfarbe darstellen
   Markiert:=m;
   if m then Font.Color:=fMarkeMarkiert else Font.Color:=fMarkeNormal;
end;

procedure TMarke.SetWert(w:integer);
begin
   Pwert:=w;
   if w<0 then begin Visible:=false; Caption:=''; end // Marke ungültig
          else begin Caption:=chr(ord('0')+w); Visible:=true; end;
end;

{ TAuswahl }
constructor TAuswahl.Create(AOwner:TComponent);
begin
   inherited Create(AOwner);
   Parent:=Hauptfenster.SpielfeldPanel;
   Shape:=bsFrame;
   OnMouseMove:=@AMouseMove; OnMouseDown:=@AMouseDown; OnClick:=@AClick; OnDblClick:=@ADblClick;
   Visible:=false;
end;

function TAuswahl.SetKoord(p:TPoint;r:integer;u:boolean):boolean;
// Auswahl ins Spielfeld setzen; p=Stelle; r=Richtung;
// u= am Rand Richtung umdrehen (Move), oder nicht (Rechtsklick=Drehen)
// Ergebnis true, wenn Setzen erfolgreich
var p1,p2:TPoint;
begin
  if Ende then exit;
  Result:=false;
  p1:=p; p2:=p+pRicht[r];
  if not PtInRect(Spielfeld.SteineMass,p2)or Spielfeld.Besetzt[p2.X,p2.Y] then // evtl umdrehen
  begin
    if not u then exit; // umdrehen nicht erlaubt
    r:=(r+2)mod 4; p2:=p+pRicht[r]; // umgedreht
    if not PtInRect(Spielfeld.SteineMass,p2)or Spielfeld.Besetzt[p2.X,p2.Y] then
    begin  // umgedreht geht auch nicht; versuche andere Richtungen
      r:=(r+1) mod 4; p2:=p+pRicht[r];
      if not PtInRect(Spielfeld.SteineMass,p2)or Spielfeld.Besetzt[p2.X,p2.Y] then
       begin
         r:=(r+2) mod 4; p2:=p+pRicht[r]; // gegensätzliche Richtung
         if not PtInRect(Spielfeld.SteineMass,p2)or Spielfeld.Besetzt[p2.X,p2.Y] then exit; // geht gar nichts
       end;
    end;
  end;
  if (p1.X>p2.X)or(p1.Y>p2.Y) then p1:=p2; // zum Zeichnen
  if (r=1)or(r=3) then SetBounds((p1.X+1)*SteinBreite,(p1.Y+1)*SteinBreite,2*SteinBreite-1,SteinBreite-1) // waagrecht
                  else SetBounds((p1.X+1)*SteinBreite,(p1.Y+1)*SteinBreite,SteinBreite-1,2*SteinBreite-1); // senkrecht
  Koord:=p; Richt:=r; Result:=true;
end;

procedure TAuswahl.AMouseMove(Sender:TObject;Shift:TShiftState;X,Y:Integer);
begin
  Hauptfenster.SpielfeldPanelMouseMove(Sender,Shift,Left+X,Top+Y);
end;

procedure TAuswahl.AMouseDown(Sender:TObject;Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
// rechte Maustaste ändert Richtung der Auswahl
var b:boolean; r,q:integer;
begin
   if Block then exit;
   if Button<>mbRight then exit;
   r:=Richt; q:=(r+1)mod 4; // nächste Richtung
   b:=false;
   while (q<>r)and(not b) do
   begin
     b:=SetKoord(Koord,q,false);
     if not b then q:=(q+1)mod 4; // fertig, wenn möglich
   end;
   if b then Richt:=q;
end;

procedure TAuswahl.AClick(Sender:TObject);
// mit Click auf Auswahl wird Stein aus Vorlage ins Spielfeld gebracht
var b:boolean; nx,ny,a1,a2:integer; p1,p2:TPoint; s:TStein;
begin
  if Block then exit;
  if (Richt=3)or(Richt=2) then // rechts oder unten
  begin // Koordinatenfeld recht oder oben
    p1:=Koord; p2:=Koord+pRicht[Richt]; // wie gegeben
  end else
  begin // andere Richtung
    p1:=Koord+pRicht[Richt]; p2:=Koord;
  end;
  a1:=Spielfeld.Marken[p1.x,p1.y].Pwert; a2:=Spielfeld.Marken[p2.X,p2.Y].Pwert;
     // Augenzahlen auf dem Spielfeld in der Auwsahl
  s:=AllSteine[a1,a2]; // Stein aufgrund Augenzahlen
  if s.inSpielfeld then exit; // Stein mit dieser Augenzahl bereits im Spielfeld
  Visible:=false; // Auswahl verbergen
  s.SetzInSpielFeld(p1,(Richt=1)or(Richt=3),Spielfeld.Marken[p1.X,p1.Y].Pwert=s.Augen.X);
  Spielfeld.MarkenUnMarkiert;
  Hauptfenster.AuslagePanel.Repaint; Spielfeld.Panel.Repaint;
  b:=false;
  for nx:=0 to Spielfeld.SteineMass.Width-1 do for ny:=0 to Spielfeld.SteineMass.Height-1 do
    if not Spielfeld.Besetzt[nx,ny] then begin b:=true; break end; // es gibt noch freie Stellen
  if not b then Hauptfenster.SpielFertig; // keine freie Stelle mehr: Spielfeld voll: Spiel zu Ende
end;

procedure TAuswahl.ADblClick(Sender:TObject);
// verlegt vorhandenen Stein an andere Stelle
var a1,a2:integer; s:TStein; p1,p2:TPoint;
begin
  if Block then exit;
  p1:=Koord; p2:=Koord+pRicht[Richt];
  a1:=Spielfeld.Marken[p1.x,p1.y].Pwert; a2:=Spielfeld.Marken[p2.X,p2.Y].Pwert;
     // Augenzahlen auf dem Spielfeld in der Auswahl
  s:=AllSteine[a1,a2]; // Stein aufgrund Augenzahlen
  if not s.inSpielfeld then exit; // Stein nicht vorhanden
  s.EntferneAusSpielfeld;
  AnzRueck:=AnzRueck+1;
  AClick(Sender);
end;

{ TStein }
constructor TStein.Create(AOwner:TComponent;a:TPoint);
begin
  Augen:=a;
  inAuslage:=false; KoordSpielfeld:=pNone; inLoesung:=false;
end;

procedure TStein.Zeichne(c:TCanvas;x,y:integer;w,r:boolean);
const // für einen Stein verwendete Farben
  fStein=clBlack; fSteinRand=clSilver; fSteinLinie=clGray; lBreite=1;
  fSteinSchatten=$00a8b8b8; fSteinPunkt=clWhite;
var d,p,ws,a1,a2:integer; e:array[0..3] of TPoint;
    deckel:array[0..3] of TPoint; mlinie1,mlinie2:TPoint;
procedure ZeichnePunkt(c:TCanvas;x,y:integer);
  var d:integer;
  begin
     d:=SteinBreite div 10; // Durchmesser 1/10 der Steingröße
     c.Pen.Width:=lBreite; c.Pen.Color:=fSteinSchatten;
     c.Brush.Color:=fSteinPunkt;
     c.Ellipse(x-d+1,y-d,x+d+1,y+d);
     c.Ellipse(x-d,y-d,x+d,y+d);
  end;
procedure ZeichneAuge(c:TCanvas;p1,p2:TPoint;a:integer);
  var xl,xm,xr,yo,ym,yu:integer;
  begin
     xm:=(p2.X+p1.X)div 2; ym:=(p2.Y+p1.Y)div 2;
     xl:=(xm+p1.X)div 2; xr:=(p2.X+xm)div 2;
     yo:=(ym+p1.Y)div 2; yu:=(p2.Y+ym)div 2;
     if a in [1,3,5,7,9] then ZeichnePunkt(c,xm,ym);
     if a in [2,4,5,6,7,8,9] then
        begin ZeichnePunkt(c,xl,yo); ZeichnePunkt(c,xr,yu) end;
     if a in [3,4,5,6,7,8,9] then
        begin ZeichnePunkt(c,xr,yo); ZeichnePunkt(c,xl,yu) end;
     if a in [6,7,8,9] then
        begin ZeichnePunkt(c,xm,yo); ZeichnePunkt(c,xm,yu) end;
     if a in [8,9] then
        begin ZeichnePunkt(c,xl,ym); ZeichnePunkt(c,xr,ym) end;
  end;
begin
   d:=SteinBreite div 18; // halber Abstand Steine = Steinbreite/18
   p:=SteinBreite div 6; // perspektivische Verschiebung = Steinbreite/6
   if w then ws:=1 else ws:=0; // waagrecht oder senkrecht
   begin
     deckel[0]:=Point(x+d+p,y+d-p);
     deckel[1]:=Point(x+(1+ws)*SteinBreite-d+p,y+d-p);
     deckel[2]:=Point(x+(1+ws)*SteinBreite-d+p,y+(2-ws)*SteinBreite-d-p);
     deckel[3]:=Point(x+d+p,y+(2-ws)*SteinBreite-d-p);
     c.Pen.Color:=fSteinRand; c.Pen.Width:=2*lBreite;
     c.Brush.Color:=fStein;
     e[0]:=Point(x+d+p,y+d-p);
     e[1]:=Point(x+d,y+d);
     e[2]:=Point(x+d,y+(2-ws)*SteinBreite-d);
     e[3]:=Point(x+d+p,y+(2-ws)*SteinBreite-d-p);
     c.Polygon(e);
     e:=deckel;
     e[0]:=Point(x+d,y+(2-ws)*SteinBreite-d);
     e[1]:=Point(x+(1+ws)*SteinBreite-d,y+(2-ws)*SteinBreite-d);
     c.Polygon(e);
     c.Polygon(deckel);
     c.Pen.Color:=fSteinLinie; c.Pen.Width:=lBreite;
     if ws=0 then
     begin
       mlinie1:=Point(x+d+p,y+SteinBreite-d-p);
       mlinie2:=Point(x+SteinBreite-d+p,mlinie1.Y);
     end else
     begin
       mlinie1:=Point(x+d+p+Steinbreite,y+d-p);
       mlinie2:=Point(mlinie1.X,y+Steinbreite-d-p);
     end;
     c.MoveTo(mlinie1.X+(1-ws)*lBreite,ws*LBreite+mlinie1.Y);
     c.LineTo(mlinie2.X-(1-ws)*lBreite,mlinie2.Y-ws*lBreite);
     if r then begin a1:=Augen.X; a2:=Augen.Y; end
          else begin a1:=Augen.Y; a2:=Augen.X; end;
     ZeichneAuge(c,deckel[0],mlinie2,a1);
     ZeichneAuge(c,mlinie1,deckel[2],a2);
   end;
end;

function TStein.inSpielfeld:boolean;
begin result:=(KoordSpielfeld.X>=0)and(KoordSpielfeld.Y>=0); end;

procedure TStein.EntferneAusSpielfeld;
var p1,p2:TPoint;
begin
   p1:=KoordSpielfeld; if WaagrSpielfeld then p2:=p1+pRechts else p2:=p1+pUnten;
   Spielfeld.Besetzt[p1.X,p1.Y]:=false; Spielfeld.Besetzt[p2.X,p2.Y]:=false;
   Spielfeld.Marken[p1.X,p1.Y].Visible:=true; Spielfeld.Marken[p2.X,p2.Y].Visible:=true;
   Spielfeld.Steine[p1.X,p1.Y]:=nil; Spielfeld.Steine[p2.X,p2.Y]:=nil;
   KoordSpielfeld:=pNone;
end;

procedure TStein.SetzInSpielfeld(p:TPoint;w,r:boolean);
// einen Stein ins Spielfeld setzen und aus Auslage entfernen
// w Lage des Steins waag-oder senkrecht; r Lage der Augen
var p2:TPoint;
begin
  if w then p2:=p+pRechts // phys. Lage waagrecht nach rechts
       else p2:=p+pUnten; // phys. Lage senkrecht nach Unten
  if inSpielfeld then EntferneAusSpielfeld;
    // Stein ist schon an anderer Stelle im Feld (Doppelklick)
  Spielfeld.Marken[p.X,p.Y].Visible:=false; Spielfeld.Marken[p2.X,p2.Y].Visible:=false;
  Spielfeld.Besetzt[p.X,p.Y]:=true; Spielfeld.Besetzt[p2.X,p2.Y]:=true;
  Spielfeld.Steine[p.X,p.Y]:=Self; Spielfeld.Steine[p2.X,p2.y]:=Self;
  KoordSpielfeld:=p; WaagrSpielfeld:=w; AugenSpielfeld:=r;
  inAuslage:=false; // Stein nicht (mehr) in Auslage
end;

procedure TStein.SetzInAuslage;
// einen Stein in AuslagePanel setzen;
var p:TPoint;
begin
  p:=KoordAuslage;
  Zeichne(Hauptfenster.AuslagePanel.Canvas,(p.X+1)*SteinBreite,(2*p.Y+1)*SteinBreite,false,true);
  inAuslage:=true;
end;


end.

